﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Linq;
using System.Security.Policy;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.Bandits.GrammarPolicies {
  public class TDPolicy : GrammarPolicy {

    private readonly HashSet<string> done;
    private readonly Dictionary<string, double> v;
    private EpsGreedyPolicy epsGreedy;

    public TDPolicy(IProblem problem, bool useCanonicalRepresentation = false)
      : base(problem, useCanonicalRepresentation) {
      this.done = new HashSet<string>();
      this.v = new Dictionary<string, double>();
      this.epsGreedy = new EpsGreedyPolicy(problem, useCanonicalRepresentation, 0.1);
    }

    public override bool TrySelect(Random random, ReadonlySequence curState, IEnumerable<ReadonlySequence> afterStates, out ReadonlySequence selectedState) {
      // only select states that are not yet done
      afterStates = afterStates.Where(a => !done.Contains(CanonicalState(a.ToString()))).ToArray();
      if (!afterStates.Any()) {
        // fail because all follow states have already been visited => also disable the current state
        done.Add(CanonicalState(curState.ToString()));
        selectedState = null;
        return false;
      }

      //return epsGreedy.TrySelect(random, curState, afterStates, out selectedState);

      var bestQ = double.NegativeInfinity;
      selectedState = null;
      foreach (var state in afterStates) {
        // try each state at least once
        if (GetTries(state) == 0) {
          selectedState = state;
          return true;
        }
        var q = V(state);
        if (q > bestQ) {
          bestQ = q;
          selectedState = state;
        }
      }

      return true;
    }

    private double V(ReadonlySequence state) {
      var s = CanonicalState(state.ToString());
      if (v.ContainsKey(s)) return v[s];
      else return 0.0;
    }

    public override void UpdateReward(IEnumerable<ReadonlySequence> stateTrajectory, double reward) {
      base.UpdateReward(stateTrajectory, reward);
      epsGreedy.UpdateReward(stateTrajectory, reward);
      // the last state could be terminal
      var lastState = stateTrajectory.Last();
      if (lastState.IsTerminal) done.Add(CanonicalState(lastState.ToString()));

      v[CanonicalState(lastState.ToString())] = V(lastState) + 1.0 / GetTries(lastState) * (reward - V(lastState));

      foreach (var p in stateTrajectory.Zip(stateTrajectory.Skip(1), Tuple.Create).Reverse()) {
        var cur = p.Item1;
        var next = p.Item2;

        v[CanonicalState(cur.ToString())] = V(cur) + 1.0 / GetTries(cur) * (V(next) - V(cur));
        //v[CanonicalState(cur.ToString())] = V(cur) + 0.1 * (V(next) - V(cur));
      }

    }

    public override double GetValue(ReadonlySequence state) {
      return V(state);
    }

    public void Reset() {
      base.Reset();
      epsGreedy.Reset();
      v.Clear();
      done.Clear();
    }
  }
}
